"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getUnsupportedEncoding = exports.unsupportedEncodingKey = exports.getPrototypeEncoder = exports.getPrototypeDecoder = exports.NullPrototypeEncoding = exports.ArrayEncoding = exports.ObjectEncoding = exports.encodeObject = exports.decodeObject = exports.nullPlaceholder = void 0;
const utils_1 = require("../utils");
exports.nullPlaceholder = {};
function getAllOwnKeys(obj, onlyEnumerable) {
    const keys = Reflect.ownKeys(obj);
    if (onlyEnumerable) {
        return keys.filter(x => Object.prototype.propertyIsEnumerable.call(obj, x));
    }
    return keys;
}
function decodeObject(target, input, ctx) {
    let stringKeys = input;
    if (Array.isArray(input)) {
        let symbolKeys;
        [stringKeys, symbolKeys] = input;
        for (const [key, value] of Object.entries(symbolKeys)) {
            target[ctx.decode(key)] = ctx.decode(value);
        }
    }
    for (const [key, value] of Object.entries(stringKeys)) {
        target[key] = ctx.decode(value);
    }
    return target;
}
exports.decodeObject = decodeObject;
function encodeObject(input, ctx, alsoNonEnumerable, explicitlyInclude = []) {
    const strKeyObject = {};
    let symbKeyObject;
    for (const key of getAllOwnKeys(input, !alsoNonEnumerable)) {
        const value = input[key];
        if (typeof key === "symbol") {
            symbKeyObject !== null && symbKeyObject !== void 0 ? symbKeyObject : (symbKeyObject = {});
            symbKeyObject[ctx.encode(key)] = ctx.encode(value);
        }
        else {
            strKeyObject[key] = ctx.encode(value);
        }
    }
    for (const key of explicitlyInclude) {
        if (!(key in strKeyObject) && key in input) {
            strKeyObject[key] = ctx.encode(input[key]);
        }
    }
    if (symbKeyObject) {
        return [strKeyObject, symbKeyObject];
    }
    ctx._isImplicit = true;
    return strKeyObject;
}
exports.encodeObject = encodeObject;
exports.ObjectEncoding = {
    prototypes: [Object.prototype],
    key: (0, utils_1.getLibraryString)("object"),
    encode(input, ctx) {
        return encodeObject(input, ctx, false);
    },
    decoder: {
        create(encodedValue, ctx) {
            return {};
        },
        init(target, encoded, ctx) {
            decodeObject(target, encoded, ctx);
        }
    }
};
function encodeAsSparseArray(input, ctx) {
    // Sparse arrays are serialized like objects.
    const result = encodeObject(input, ctx, false);
    ctx._isImplicit = false;
    return result;
}
exports.ArrayEncoding = {
    key: (0, utils_1.getLibraryString)("array"),
    prototypes: [Array.prototype],
    encode(input, ctx) {
        const keys = Object.keys(input);
        let isSparseCanFalseNegative = input.length !== keys.length;
        if (isSparseCanFalseNegative) {
            return encodeAsSparseArray(input, ctx);
        }
        // The array still might be sparse, even after that check.
        const newArray = [];
        for (let i = 0; i < keys.length; i++) {
            if (i !== +keys[i]) {
                return encodeAsSparseArray(input, ctx);
            }
            newArray.push(ctx.encode(input[i]));
        }
        ctx._isImplicit = true;
        return newArray;
    },
    decoder: {
        create(encodedValue, ctx) {
            return [];
        },
        init(target, input, ctx) {
            if (!Array.isArray(input)) {
                // Decode similarly to objects
                decodeObject(target, input, ctx);
                return;
            }
            for (let i = 0; i < input.length; i++) {
                target[i] = ctx.decode(input[i]);
            }
        }
    }
};
exports.NullPrototypeEncoding = {
    key: (0, utils_1.getLibraryString)("null"),
    encode: getPrototypeEncoder(null),
    decoder: getPrototypeDecoder(null),
    prototypes: [exports.nullPlaceholder]
};
function getPrototypeDecoder(proto) {
    return {
        init: exports.ObjectEncoding.decoder.init,
        create(encodedValue, ctx) {
            return Object.create(proto);
        }
    };
}
exports.getPrototypeDecoder = getPrototypeDecoder;
function getPrototypeEncoder(proto) {
    return (input, ctx) => {
        const result = encodeObject(input, ctx, false);
        ctx._isImplicit = false;
        return result;
    };
}
exports.getPrototypeEncoder = getPrototypeEncoder;
exports.unsupportedEncodingKey = (0, utils_1.getLibraryString)("unsupported");
function getUnsupportedEncoding(...protos) {
    return {
        key: exports.unsupportedEncodingKey,
        prototypes: protos,
        encode(input, ctx) {
            ctx.metadata = (0, utils_1.getClassName)(input);
            return 0;
        },
        decoder: {
            create(encodedValue, ctx) {
                return undefined;
            }
        }
    };
}
exports.getUnsupportedEncoding = getUnsupportedEncoding;

//# sourceMappingURL=basic.js.map