"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.defaultConfig = exports.Szr = void 0;
const utils_1 = require("./utils");
const szr_representation_1 = require("./szr-representation");
const basic_1 = require("./encodings/basic");
const scalar_1 = require("./encodings/scalar");
const errors_1 = require("./errors");
const binary_1 = require("./encodings/binary");
const collections_1 = require("./encodings/collections");
const built_in_1 = require("./encodings/built-in");
const encoding_constructors_1 = require("./encoding-constructors");
const builtinEncodings = [
    basic_1.ObjectEncoding,
    basic_1.ArrayEncoding,
    basic_1.NullPrototypeEncoding,
    (0, scalar_1.createFundamentalObjectEncoding)(Number),
    (0, scalar_1.createFundamentalObjectEncoding)(Boolean),
    (0, scalar_1.createFundamentalObjectEncoding)(String),
    scalar_1.dateEncoding,
    scalar_1.regexpEncoding,
    ...binary_1.typedArrayEncodings,
    binary_1.ArrayBufferEncoding,
    collections_1.MapEncoding,
    collections_1.SetEncoding,
    ...built_in_1.errorEncodings
];
const builtinUnsupportedTypes = [
    WeakMap.prototype,
    WeakSet.prototype,
    Function.prototype
];
/**
 * The class used to encode and decode things in the szr format.
 */
class Szr {
    constructor(config) {
        this._config = exports.defaultConfig;
        this._keyToEncoding = new Map();
        this._symbToEncoding = new Map();
        this._tempSymbEncoding = new Map();
        this._protoEncodingCache = new WeakMap();
        this._protoEncodings = [];
        this._config = (0, utils_1.defaultsDeep)({}, config, this._config);
        const unsupportedEncoding = (0, basic_1.getUnsupportedEncoding)(...builtinUnsupportedTypes, ...this._config.unsupported);
        this.addEncoding(...builtinEncodings, unsupportedEncoding, ...this._config.encodings);
    }
    _buildEncodingCache() {
        this._protoEncodingCache = new WeakMap();
        for (const encoding of this._protoEncodings) {
            for (const proto of encoding.prototypes) {
                this._protoEncodingCache.set(proto, encoding);
            }
        }
    }
    addEncoding(...encoders) {
        for (const encSpecifier of encoders) {
            const encoding = (0, encoding_constructors_1.getFullEncoding)(encSpecifier);
            if (this._keyToEncoding.get(encoding.key)) {
                throw new errors_1.SzrError(`Encoding with the key '${encoding.key}' already exists.`);
            }
            if ("prototypes" in encoding) {
                this._protoEncodings.push(encoding);
            }
            else {
                this._symbToEncoding.set(encoding.symbol, encoding);
            }
            this._keyToEncoding.set(encoding.key, encoding);
        }
        this._buildEncodingCache();
    }
    _findEncodingForObject(obj) {
        var _a;
        if (Array.isArray(obj)) {
            return basic_1.ArrayEncoding;
        }
        let foundEncoding;
        for (let proto = obj;; proto = (_a = Object.getPrototypeOf(proto)) !== null && _a !== void 0 ? _a : basic_1.nullPlaceholder) {
            const cached = this._protoEncodingCache.get(proto);
            if (cached !== undefined) {
                foundEncoding = cached;
                break;
            }
            if (proto === basic_1.nullPlaceholder) {
                throw new errors_1.SzrError("FindEncodingForObject got stuck. Internal Error.");
            }
        }
        foundEncoding !== null && foundEncoding !== void 0 ? foundEncoding : (foundEncoding = basic_1.ObjectEncoding);
        if (!this._protoEncodingCache.has(obj)) {
            this._protoEncodingCache.set(obj, foundEncoding);
        }
        return foundEncoding;
    }
    _findEncodingForSymbol(symb) {
        var _a;
        let encoding = (_a = this._symbToEncoding.get(symb)) !== null && _a !== void 0 ? _a : this._tempSymbEncoding.get(symb);
        if (encoding == null) {
            encoding = {
                symbol: symb,
                key: szr_representation_1.unrecognizedSymbolKey,
                metadata: (0, utils_1.getSymbolName)(symb) || `#${this._symbToEncoding.size + 1}`
            };
            this._tempSymbEncoding.set(symb, encoding);
        }
        return encoding;
    }
    _findEncodingByKeyValue(input, encodingKey) {
        if (encodingKey != null) {
            const encoding = this._keyToEncoding.get(encodingKey);
            return encoding;
        }
        if (Array.isArray(input)) {
            return basic_1.ArrayEncoding;
        }
        return basic_1.ObjectEncoding;
    }
    _checkInputValid(input) {
        let reason = "";
        let versionInfo = "";
        if (!Array.isArray(input)) {
            reason = "input is not array";
        }
        else {
            const header = input === null || input === void 0 ? void 0 : input[0];
            if (!header) {
                reason = "no header element";
            }
            else if (!Array.isArray(header)) {
                reason = "header element is not array";
            }
            else {
                versionInfo = header[0];
                if (versionInfo == null) {
                    reason = "no version info";
                }
                else if (typeof versionInfo !== "string") {
                    reason = "version is not string";
                }
                else if (+versionInfo !== parseInt(versionInfo)) {
                    reason = "version is not numeric";
                }
                else if (versionInfo !== utils_1.version) {
                    throw new errors_1.SzrError(`Input was encoded using version ${versionInfo}, but szr is version ${utils_1.version}. Set skipValidateVersion to allow this.`);
                }
                else if (!Array.isArray(header[1])) {
                    reason = "no encoding keys or encoding keys not an array";
                }
                else if (typeof header[2] !== "object" || !header[1]) {
                    reason = "no encoding data or encoding data is not an object";
                }
                else if (typeof header[3] !== "object" || !header[2]) {
                    reason = "no custom metadata or custom metadata is not an object";
                }
                else if (input.length === 1) {
                    reason = "input must have at least 2 elements";
                }
            }
        }
        if (reason) {
            throw new errors_1.SzrError(`Input is not szr-encoded: ${reason}`);
        }
    }
    decode(input) {
        const tryScalar = (0, szr_representation_1.tryDecodeScalar)(input);
        if (tryScalar !== szr_representation_1.noResultPlaceholder)
            return tryScalar;
        input = input;
        this._checkInputValid(input);
        const header = input === null || input === void 0 ? void 0 : input[0];
        const [v, encodingKeys, encodingSpec, metadata] = header;
        const targetArray = Array(input.length - 1);
        const needToInit = new Map();
        const ctx = {
            decode: null,
            metadata: undefined
        };
        for (const encodingKey of encodingKeys) {
            if (!this._keyToEncoding.has(encodingKey) && encodingKey !== szr_representation_1.unrecognizedSymbolKey) {
                throw new errors_1.SzrError(`Encoding with key '${encodingKey}' not found. Szr wasn't configured correctly.`);
            }
        }
        for (let i = 1; i < input.length; i++) {
            const encodingIndex = encodingSpec[i];
            const encodingKey = encodingKeys[encodingIndex];
            let cur = input[i];
            if (encodingKey === szr_representation_1.unrecognizedSymbolKey) {
                targetArray[i] = (0, utils_1.getUnrecognizedSymbol)(metadata[i]);
                continue;
            }
            if (encodingKey == null && typeof cur === "string") {
                targetArray[i] = cur;
                continue;
            }
            const encoding = this._findEncodingByKeyValue(cur, encodingKey);
            if ("symbol" in encoding) {
                targetArray[i] = encoding.symbol;
                continue;
            }
            ctx.metadata = metadata[i];
            targetArray[i] = encoding.decoder.create(cur, ctx);
            if (encoding.decoder.init) {
                needToInit.set(i, encoding);
            }
        }
        ctx.decode = (value) => {
            const decodedPrimitive = (0, szr_representation_1.tryDecodeScalar)(value);
            if (decodedPrimitive !== szr_representation_1.noResultPlaceholder)
                return decodedPrimitive;
            return targetArray[value];
        };
        for (let key of needToInit.keys()) {
            const encoding = needToInit.get(key);
            ctx.metadata = metadata[key];
            encoding.decoder.init(targetArray[key], input[key], ctx);
        }
        return targetArray[1];
    }
    encode(root) {
        try {
            const tryScalar = (0, szr_representation_1.tryEncodeScalar)(root);
            if (tryScalar !== szr_representation_1.noResultPlaceholder)
                return tryScalar;
            const encodingSpec = {};
            const metadata = {};
            const encodingKeys = new Map();
            const header = [];
            const szrRep = [header];
            const objectToRef = new Map();
            const cacheEncodingIndex = (key) => {
                let encodingIndex = encodingKeys.get(key);
                if (encodingIndex == null) {
                    encodingIndex = encodingKeys.size;
                    encodingKeys.set(key, encodingIndex);
                }
                return encodingIndex;
            };
            const ctx = {
                metadata: undefined,
                encode(value) {
                    const tryScalar = (0, szr_representation_1.tryEncodeScalar)(value);
                    if (tryScalar !== szr_representation_1.noResultPlaceholder)
                        return tryScalar;
                    let existingRef = objectToRef.get(value);
                    if (!existingRef) {
                        existingRef = createNewRef(value);
                    }
                    return existingRef;
                }
            };
            const createNewRef = (value) => {
                const index = szrRep.length;
                const ref = `${index}`;
                objectToRef.set(value, ref);
                if (typeof value === "string") {
                    szrRep.push(value);
                    return ref;
                }
                szrRep.push(0);
                if (typeof value === "symbol") {
                    const encoding = this._findEncodingForSymbol(value);
                    encodingSpec[index] = cacheEncodingIndex(encoding.key);
                    if (encoding.metadata) {
                        metadata[index] = encoding.metadata;
                    }
                    return ref;
                }
                const encoding = this._findEncodingForObject(value);
                const oldMetadata = ctx.metadata;
                const szed = encoding.encode(value, ctx);
                if (!ctx._isImplicit) {
                    encodingSpec[index] = cacheEncodingIndex(encoding.key);
                }
                ctx._isImplicit = false;
                if (ctx.metadata !== undefined) {
                    metadata[index] = ctx.metadata;
                }
                ctx.metadata = oldMetadata;
                szrRep[index] = szed;
                return ref;
            };
            ctx.encode(root);
            header.push(utils_1.version, [...encodingKeys.keys()], encodingSpec, metadata);
            return szrRep;
        }
        finally {
            this._tempSymbEncoding.clear();
        }
    }
}
exports.Szr = Szr;
exports.defaultConfig = {
    encodings: [],
    unsupported: []
};

//# sourceMappingURL=szr.js.map